# datetimelingo.py
# DateTimeLingo extends datetime with natural language elements
# Date created: 12 Oct 2007
# Date updated: 1/15/08
#
# Authors: Brian Gershon and Derek Hoshiko
# Copyright 2007-2008 Web Collective, Inc.
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from datetime import datetime, timedelta
import time
import re

STR_TODAY = 'today'
STR_YESTERDAY = 'yesterday'
STR_TOMORROW = 'tomorrow'

STRFTIME_TO_USER_FRIENDLY_TOKENS = {
   'm' : 'mm',
   'm' : 'm',
   'd' : 'dd',
   'd' : 'd',
   'Y' : 'yyyy',
   'y' : 'yy',
}

USER_FRIENDLY_TOKENS_TO_STRFTIME = {
   'mm'   : 'm',
   'm'    : 'm',
   'dd'   : 'd',
   'd'    : 'd',
   'yyyy' : 'Y',
   'yy'   : 'y',
}

SINGLE_DIGITS = [1,2,3,4,5,6,7,8,9,0]

WEEKDAY_MAPPING = [
    { 'weekday_number':0, 'abbrev':'mon', 'name':'monday' },
    { 'weekday_number':1, 'abbrev':'tue', 'name':'tuesday' },
    { 'weekday_number':2, 'abbrev':'wed', 'name':'wednesday' },
    { 'weekday_number':3, 'abbrev':'thr', 'name':'thursday' },
    { 'weekday_number':4, 'abbrev':'fri', 'name':'friday' },
    { 'weekday_number':5, 'abbrev':'sat', 'name':'saturday' },
    { 'weekday_number':6, 'abbrev':'sun', 'name':'sunday' },
]
def getDayList():
    return_list = []
    for weekday in WEEKDAY_MAPPING:
        return_list.append(weekday['abbrev'])
        return_list.append(weekday['name'])
    return return_list
    
def _findNextWeekdayDate(weekday_string_param):
    """
        Convert weekday name or abbreviation into a weekday_number,
        then return a Date.
    """
    weekday_string = weekday_string_param.lower().strip()
    
    weekday_number = None
    for weekday in WEEKDAY_MAPPING:
        if weekday['abbrev'].find(weekday_string) != -1:
            weekday_number = weekday['weekday_number']
            continue
        if weekday['name'].find(weekday_string) != -1:
            weekday_number = weekday['weekday_number']
            continue

    if weekday_number is None:
        return None
    else:
        today = datetime.now()
        weekday_today = datetime.now().weekday()
        new_date = None
        if weekday_today == weekday_number:
            # return the date one week from now
            new_date = today + timedelta(7)
        else:
            # cycle through the days until weekday() = weekday_number
            for i in range(1,8):
                next_day = today + timedelta(i)
                if next_day.weekday() == weekday_number:
                    new_date = today + timedelta(i)
        return new_date

def createDate(datestring):
    """ Return a 'datetime' from a string, or raise an error.
        String should be:
            - in a standard date string format
            - a constant: 'now', 'tomorrow', 'yesterday'
                or day of week name or abbreviation.
    """

    # reg expressions to figure out various dates

    # US: mm/dd/yyyy and m/d/yyyy
    strRE = r'(0*[1-9]|1[012])[- /.](0*[1-9]|[12][0-9]|3[01])[- /.](19|20)\d\d'
    re_date_usa = re.compile(strRE)

    # EUROPE: yyyy/mm/dd and yyyy/m/d
    strRE = r'(19|20)\d\d[- /.](0*[1-9]|1[012])[- /.](0*[1-9]|[12][0-9]|3[01])'
    re_date_eur = re.compile(strRE)
    
    # Ambiguous: xx/yy/zz   is xx the year (European)? or is zz (US)?
    strRE = r'[0-9]*\d[- /.](0*[1-9]|1[012])[- /.][0-9]*\d'
    re_date_ambiguous = re.compile(strRE)

    # Ambiguous: xx/yy/zz   is xx the year (European)? or is zz (US)?
    # e.g. 4/20/08
    strRE = r'[0-9]*\d[- /.]([0-9]*\d)[- /.]0[0-9]*\d'
    re_date_ambiguous2 = re.compile(strRE)

    dt = datestring.lower().strip()
    
    dtNow = datetime.now()
    if dt == STR_TODAY:
        return dtNow
    elif dt == STR_YESTERDAY:
        return dtNow - timedelta(1)
    elif dt == STR_TOMORROW:
        return dtNow + timedelta(1)
    elif dt in getDayList():
        return _findNextWeekdayDate(dt)
    elif re_date_usa.search(dt) is not None:
        return datetime(*time.strptime(dt, "%m/%d/%Y")[0:5])
    elif re_date_eur.search(dt) is not None:
        return datetime(*time.strptime(dt, "%Y/%m/%d")[0:5])
    elif re_date_ambiguous.search(dt) is not None or re_date_ambiguous2.search(dt) is not None:
        # FIXME: Should this default be a value in a property sheet,
        #        Or should we use the system default format?

        # default to US
        try:
            new_date = datetime(*time.strptime(dt, "%m/%d/%y")[0:5])
            return new_date
        except ValueError:
            pass

        # try European next (and if not, intentionally error out)
        try:
            new_date = datetime(*time.strptime(dt, "%y/%m/%d")[0:5])
            return new_date
        except ValueError:
            pass

        errStr = '"%s" date format not recognized in createDate().'
        raise ValueError, errStr % datestring
    else:
        errStr = '"%s" date format not recognized in createDate().'
        raise ValueError, errStr % datestring

def displayDate(dtInput, strFormat="mm/dd/yyyy"):
    """ Get a string suitable for display of a datetime object using lingo.  If
    the date is today, tomorrow, or yesterday, displayDate returns 
    such corresponding strings.  
    
    If a lingo translation is not available, displayDate will return a string 
    formatted as specified by strFormat.  Accepts user-friendly canned formats 
    as follows.
        U.S.
                 mm/dd/yyyy   10/01/1976  <---- Default
                 m/d/yyyy     10/1/1976
TODO:            mm/dd/yy     10/01/76
TODO:            m/d/yy       10/1/76
TODO:            mm/dd        10/01
TODO:            m/d          10/1
        Europe
                 yyyy/mm/dd   1976/10/01
                 yyyy/m/d     1976/10/1
TODO:            dd/mm/yy     01/10/07
TODO:            d/m/yy       1/10/07
        Other
TODO:            dd Mmm yyyy  01 Oct 1976
TODO:            dd Mmm       01 Oct            
    """
    # Validate input
    if dtInput is None:
        return ""
    if type(dtInput) != datetime:
        raise ValueError, "dtInput must be None or a datetime"
        
    dtInputDate = dtInput.date()
    dtTodayDate = datetime.now().date()
    
    # Lingo
    if dtInputDate == dtTodayDate:
        return STR_TODAY
    elif dtInputDate == dtTodayDate - timedelta(1):
        return STR_YESTERDAY
    elif dtInputDate == dtTodayDate + timedelta(1):
        return STR_TOMORROW
    
    # Format date
    # TODO: Factor out strftime translations into a separate function or even
    # a class or even a module.
    if '/' in strFormat:
        strftimeTokens = []
        formatTokens = strFormat.split('/')
        for token in formatTokens:
            if token == 'm':
                newToken = str(dtInputDate.month)
            elif token == 'd':
                newToken = str(dtInputDate.day)
            else:
                newToken = "%" + USER_FRIENDLY_TOKENS_TO_STRFTIME[token]
            strftimeTokens.append(newToken)
        newFormatString = '/'.join(strftimeTokens)
        return dtInputDate.strftime(newFormatString)
    return ""
