# effort.py
# Effort provides a class that represents effort as either a scalar or a 
# vector, with some provision for ambiguity in a succinct and well-defined
# format.  Please note, however, that the vector format is the intended use of
# this class.
# Date created: 01 Feb 2008
#
# Author: Derek Hoshiko
# Copyright 2007-2008 Web Collective, Inc.
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
#
# D O C U M E N T A T I O N
# 
# Scalar Format -
# In scalar format, effort may be a single number or text string, and each 
# object contains a single value.  Scalar format is offered as a convenience
# in data entry from a user, and for future support of other methodologies such 
# as ranges of numbers, like 1-5, or string sequences like [ 'Difficult', 
# 'Moderate', 'Easy' ].
#     Single Number -
#         - Can be integer or floating point.
#         - Can be compared.
#     Text Strings -
#         - Can be any valid Python string.
#         - Can not be compared, and is offered only as a convenience.
#
# TODO: Write code to support Scalar Format - currently not supported.
#
# Vector Format
# In vector format, effort becomes a formatted text string with the following 
# syntax, and with the ability to perform mathematical operations on effort
# objects.
#     Grammar -
#         <Effort>
#         <Effort> -> <Effort - Vector>
#         <Effort - Vector> -> (<effort>)
#         <effort> -> <known effort><ambiguous effort>
#         <ambiguous effort> -> ? | + | NULL
#         <known effort> -> <range> | <integer> | NULL
#         <range> -> <integer>-<integer>
#         <integer> -> N, N in [0, EFFORT_LIMIT)
#
#     Formats -
#         Unknown Value - (?)
#         Known Value - (n)
#         Minimum Value - (n+)
#         Estimated Value - (n?)
#         Known Range - (n-m)
#         Minimum Range - (n-m+)
#         Estimated Range - (n-m?)
#         
#     Examples -
#         1. It will take eight to ten hours.
#             (8-10)
#         2. It will take two hours.
#             (2)
#         3. It will take at least three hours, but maybe more.
#             (3+)
#         4. It will take up to an hour.
#             (0-1)
#         5. It won't take hardly any time at all to complete.
#             (0)
#         6. I have no idea how long it will take.
#             (?)
#         7. It will take about ten hours, but could be more or less.
#             (10?)
#         8. It will take between one and three hours, but not exactly sure.
#             (1-3?)
#         9. It might take an hour, but could be more or less.
#             (1?)
#        10. It might take an hour or two, but could take longer.
#             (1-2+)
#        11. It will take some time to do.  Interpreted as (0+).
#             (+)
#
#     Operations -
#         Addition -
#             TODO: Implement addition (and probably make better definitions for
#                   the rules.
#             i) Known values and ranges are additive.
#                 (0) + (2) = (2)
#                 (8-10) + (3) = (11-13)
#                 (1-3) + (2-6) = (3-9)
#             ii) Minimum values and ranges are cumulative.
#                 (1) + (3+) = (4+)
#                 (1-3+) + (4) = (5-7+)
#             iii) Estimated values and ranges are cumulative.
#                 (2) + (6?) = (8?)
#                 (2-3) + (4-5?) = (6-8?)
#             iv) Minimum values and ranges take precedence.
#                 (1-2?) + (2-3+) = (3-5+)
#             v) Additive identity is (0)
#
#         Subtraction -
#             TODO: Implement subtraction (and probably make better definitions 
#                   for the rules.
#             i) Known values and ranges.
#                 (0) - (2) = (0)
#                 (8-10) - (3) = (5-7)
#                 (3) - (8-10) = (0)
#                 (2-6) - (1-3) = (1-3)
#             ii) Minimum values and ranges.
#                 (1-3+) - (1) = (0-3+)
#                 (3+) - (1) = (2+)
#             iii) Estimated values and ranges are cumulative.
#                 (2) - (6?) = (0)
#                 (2-3) - (4-5?) = (0)
#                 (6?) - (2) = (4?)
#                 (4-5?) - (2-3) = (2)
#             iv) Minimum values and ranges have no effect.
#                 (5-9?) - (1-2+) = (4-7?)
#                 (5-9+) - (1-2?) = (4-7+)
#             v) Additive identity is (0)
#                 (1-3) - (0) = (1-3)
#                 (2) - (0) = (2)
#             vi) Minimum values as subtrahends do strange things.
#                 (5) - (3+) = (0-2)
#                 (4) - (1-3+) = (0-3+)
#
#         Scalar Multiplication -
#             Unknown Value - k(?) = (?)
#             Known Value - k(n) = (k*n)
#             Minimum Value - k(n+) = (k*n+)
#             Estimated Value - k(n?) = (k*n?)
#             Known Range - k(n-m) = (k*n + k*m)
#             Minimum Range - k(n-m+) = (k*n-k*m+)
#             Estimated Range - k(n-m?) = (k*n-k*m?)

EFFORT_LIMIT = 1024

class Effort:
    def __init__(self, effort=None):
        self.Set(effort)
    
    def Set(self, effort):
        self._userInput = str(effort)
        return self.Build()
    
    def DisplayEffort(self):
        """ Return the user-friendly string for this Effort. """
        strVectorFormat = "(%s)"
        value = self.GetEffort('value')
        estimated = self.GetEffort('estimated')
        minimum = self.GetEffort('minimum')
        range = self.GetEffort('range')
       
        strEffort = None
        if value != False:
            strEffort = str(value)
        elif minimum != False:
            strEffort = minimum + '+'
        elif estimated != False:
            strEffort = estimated + '?'
        elif range != False:
            strEffort = '-'.join(range)

        if strEffort:
            return strVectorFormat % strEffort
        return ""
    
    def GetEffort(self, field=None):
        """ Get the effort dictionary or portions thereof, based on whether
        'field' is present or not. No interpretation is made on the state of
        the effort object; data is returned in its raw format. """
        try:
            if field:
                return self._dictoEffort[field]
            return self._dictoEffort
        except AttributeError:
            if field:
                return ""
            return {}
    
    def GetLow(self):
        """ Interpretation method: Return the minimum value of this Effort, 
        regardless of whether it is Known, Minimum, or Estimated. """
        # TODO: Implement and test as described in the comment.
        return self.GetEffort('minimum')
        
    def Build(self):
        self._vectorFormat = self.VectorFormat(self._userInput)
        self._isSingleNumber = self.IsSingleNumber(self._userInput)
        self._isTextString = self.IsTextString(self._userInput)
        if self._vectorFormat:
            minimum = self.Minimum(self._vectorFormat)
            estimated = self.Estimated(self._vectorFormat)
            if minimum:
                range = self.Range(minimum)
            elif estimated:
                range = self.Range(estimated)
            else:
                range = self.Range(self._vectorFormat)
            if range:
                value = False
            else:
                value = self.Value(self._vectorFormat)
            self._dictoEffort = {
                'minimum' : minimum,
                'estimated' : estimated,
                'range' : range,
                'value' : value,
            }
            return self._dictoEffort
        return False

    def Value(self, integer):
        """ Stage Four: Check an integer string 'integer' to determine whether
        it is a proper string representing a number, and the value is less than
        EFFORT_LIMIT, then return 'integer'.  Return False otherwise. """
        try:
            integer = str(int(integer))
        except ValueError:
            return False
        return integer
        
    def Range(self, knownEffort):
        """Stage Three: Check a known effort formatted string 'knownEffort' 
        given without enclosing parenthesis, nor any ambiguity marker, and 
        determine whether 'knownEffort' is a range or not.  If so, return a list
        containing the two elements of the range.  If not, return False. """
        listRange = knownEffort.split('-')
        for endPoint in listRange:
            if not self.IsSingleNumber(endPoint):
                return False
        if len(listRange) == 2:
            return listRange
        return False
        
    def Estimated(self, effort):
        """ Stage Two: Check a vector formatted string given without enclosing 
        parenthesis and determine whether there is a Estimated Value or 
        Estimated Range ambiguity marker '?'.  If so, return the 'value' without 
        the ambiguity marker, otherwise return False. """
        return self._Ambiguity(effort, '?')

    def Minimum(self, effort):
        """ Stage Two: Check a vector formatted string given without enclosing 
        parenthesis and determine whether there is a Minimum Value or Minimum 
        Range ambiguity marker '+'.  If so, return the 'value' without the 
        ambiguity marker, otherwise return False. """
        return self._Ambiguity(effort, '+')

    def VectorFormat(self, value):
        """ Stage One: Check whether the string representation of 'value' is in 
        a proper vector format.  Returns 'value' without enclosing parenthesis 
        if so, and False otherwise. """
        strValue = str(value)
        if strValue.startswith('(') and strValue.endswith(')'):
            return value[1:len(value)-1]
        return False

    def IsTextString(self, value):
        """ Check the value of 'value' to determine whether it is a text string
        or not.  Returns True if so, and False otherwise. """
        return not self.VectorFormat(value)

    def IsSingleNumber(self, value):
        """ Check the value of 'value' to determine whether it is a single 
        value or not.  Returns True if so, and False otherwise. """
        singleNumberCounter = 1
        isSingleNumber = False
        try:
            int(value)
        except ValueError:
            singleNumberCounter -= 1
        try:
            float(value)
        except ValueError:
            singleNumberCounter -= 1
        if singleNumberCounter > 0:
            isSingleNumber = True
        return isSingleNumber

    def _Ambiguity(self, effort, marker):
        """ Check a vector formatted string 'effort' given without enclosing 
        parenthesis and determine whether there is an ambiguity marker 'marker'.
        If so, return the 'value' without the ambiguity marker, otherwise return 
        False. """
        if effort.endswith(marker):
            newEffort = effort[0:len(effort)-1]
            isSingleNumber = self.IsSingleNumber(newEffort)
            isRange = self.Range(newEffort)
            if isSingleNumber or isRange or newEffort == '':
                return newEffort
        return False
