# Test for DateTimeLingo
# Author: Brian Gershon and Derek Hoshiko
# Created: 12 Oct 2007
# Last updated: 1/15/08

from datetime import datetime, timedelta

# Plone specific:
from Products.CollectiveGTD.tests import base
from Products.CollectiveGTD.modules.datetimelingo import createDate, displayDate

STR_TODAY = 'today'
STR_YESTERDAY = 'yesterday'
STR_TOMORROW = 'tomorrow'

WEEKDAY_MAPPING = [
    { 'weekday_number':0, 'abbrev':'mon', 'name':'monday' },
    { 'weekday_number':1, 'abbrev':'tue', 'name':'tuesday' },
    { 'weekday_number':2, 'abbrev':'wed', 'name':'wednesday' },
    { 'weekday_number':3, 'abbrev':'thr', 'name':'thursday' },
    { 'weekday_number':4, 'abbrev':'fri', 'name':'friday' },
    { 'weekday_number':5, 'abbrev':'sat', 'name':'saturday' },
    { 'weekday_number':6, 'abbrev':'sun', 'name':'sunday' },
]

class CreateDateFunctionTests(base.CollectiveGTDTestCase):
    def afterSetUp(self):
        self.today = datetime.today().date()
        
    def testToday(self):
        datespan = createDate(STR_TODAY).date() - self.today
        errStr = "createDate did not return today's date"
        self.assertEqual(datespan.days, 0, errStr)

    def testYesterday(self):
        datespan = createDate(STR_YESTERDAY).date() - self.today
        errStr = "createDate did not return yesterday's date"
        self.assertEqual(datespan.days, -1, errStr)

    def testTomorrow(self):
        datespan = createDate(STR_TOMORROW).date() - self.today
        errStr = "createDate did not return tomorrow's date"
        self.assertEqual(datespan.days, 1, errStr)
        
    def _findNextWeekdayDate(self, weekday_number):
        """
            Monday  is weekday_number 0
            Tuesday is weekday_number 1
            Sunday  is weekday_number 6
        """
        today = datetime.now()
        weekday_today = datetime.now().weekday()
        new_date = None
        if weekday_today == weekday_number:
            # return the date one week from now
            new_date = today + timedelta(7)
        else:
            # cycle through the days until weekday() = weekday_number
            for i in range(1,8):
                next_day = today + timedelta(i)
                if next_day.weekday() == weekday_number:
                    new_date = today + timedelta(i)
        return new_date
        
    def testWeekdayNames(self):
        """
        Test all the weekdays.  If today is same day as the
        test day, return the following Monday.
        """
        for weekday in WEEKDAY_MAPPING:
            test_date = self._findNextWeekdayDate(weekday_number=weekday['weekday_number'])
        errStr = "createDate did not return the proper day (%s)" % weekday['name']
        self.failUnless(createDate(weekday['abbrev']).date() == test_date.date(), errStr)
        self.failUnless(createDate(weekday['name']).date() == test_date.date(), errStr)
        
    def testUSDateStringSingleDigits(self):
        d = createDate("9/5/2007")
        datespan = d.date() - datetime(2007, 9, 5).date()
        self.assertEqual(datespan.days, 0, \
            "createDate did not properly convert US date")

    def testUSDateStringDoubleDigits(self):
        d = createDate("09/05/2007")
        datespan = d.date() - datetime(2007, 9, 5).date()
        self.assertEqual(datespan.days, 0, \
            "createDate did not properly convert US date")

    def testEuropeanDateStringSingleDigits(self):
        d = createDate("2007/9/5")
        datespan = d.date() - datetime(2007, 9, 5).date()
        self.assertEqual(datespan.days, 0, \
            "createDate did not properly convert European date")

    def testEuropeanDateStringDoubleDigits(self):
        d = createDate("2007/09/05")
        datespan = d.date() - datetime(2007, 9, 5).date()
        self.assertEqual(datespan.days, 0, \
            "createDate did not properly convert European date")
    
    def testInvalidDate(self):
        # TODO - should beef this up with lots of crappy dates
        self.assertRaises(ValueError, createDate, "17/32/2007")

    def testDefaultsToUSDateFormat(self):
        # A date such as 4/20/08 could be two different dates
        #   depending on if it's interpreted as a US format
        #   or a European format.
        # Since this is currently a US version,
        #   let's default to assuming this is US format
        #   (rather than raise an error)
        d = createDate("4/20/08")
        datespan = d.date() - datetime(2008, 4, 20).date()
        self.assertEqual(datespan.days, 0, \
            "Ambiguous date should default to US")
        
class DisplayDateFunctionTests(base.CollectiveGTDTestCase):
    def afterSetUp(self):
        self.today = datetime.today().date()
        
    def testDateValidation(self):
        displayDate(None)   # this should not raise an error
        self.assertRaises(ValueError, displayDate, "555")
        
    def testToday(self):
        dtNow = createDate(STR_TODAY)
        strResult = displayDate(dtNow)
        errStr = "displayDate did not give a proper string for today's date"
        self.assertEqual(strResult, STR_TODAY, errStr)

    def testYesterday(self):
        dtNow = createDate(STR_YESTERDAY)
        strResult = displayDate(dtNow)
        errStr = "displayDate did not give a proper string for yesterday's date"
        self.assertEqual(strResult, STR_YESTERDAY, errStr)

    def testTomorrow(self):
        dtNow = createDate(STR_TOMORROW)
        strResult = displayDate(dtNow)
        errStr = "displayDate did not give a proper string for tomorrow's date"
        self.assertEqual(strResult, STR_TOMORROW, errStr)
        
    def testUSDateStringSingleDigits(self):
        dt = createDate("9/5/2007")
        strResult = displayDate(dt, 'm/d/yyyy')
        errStr = "incorrect date string for given date format: m/d/yyyy"
        self.assertEqual(strResult, '9/5/2007', errStr)

    def testUSDateStringDoubleDigits(self):
        dt = createDate("09/05/2007")
        strResult = displayDate(dt, 'mm/dd/yyyy')
        errStr = "incorrect date string for given date format: mm/dd/yyyy"
        self.assertEqual(strResult, '09/05/2007', errStr)

    def testEuropeanDateStringSingleDigits(self):
        dt = createDate("2007/9/5")
        strResult = displayDate(dt, 'yyyy/m/d')
        errStr = "incorrect date string for given date format: yyyy/m/d"
        self.assertEqual(strResult, '2007/9/5', errStr)

    def testEuropeanDateStringDoubleDigits(self):
        dt = createDate("2007/09/05")
        strResult = displayDate(dt, 'yyyy/mm/dd')
        errStr = "incorrect date string for given date format: yyyy/mm/dd"
        self.assertEqual(strResult, '2007/09/05', errStr)

if __name__ == '__main__':
    unittest.main()
    
def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(CreateDateFunctionTests))
    suite.addTest(makeSuite(DisplayDateFunctionTests))
    return suite

